/*
** Module   : CommonWindow.java
** Abstract : Define all window-related APIs available via a window handle
**            instance.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 CA  20130112 Created initial version.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
 
/**
 * Provide access to all window-related APIs which can be accessed via a 
 * handle instance.
 */
public interface CommonWindow
{
   /**
    * Get icon file name previously loaded by {@link #loadIcon(character)}.
    * 
    * @return  icon file name.
    */
   public character getIcon();

   /**
    * Get window maximal height.
    * 
    * @return  maximal height.
    */
   public decimal getMaxHeight();

   /**
    * Gets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @return current value of MAX-HEIGHT-CHARS attribute
    */
   public decimal getMaxHeightChars();

   /**
    * Get window maximal width.
    * 
    * @return  maximal width.
    */
   public decimal getMaxWidth();

   /**
    * Gets the MAX-WIDTH-CHARS writable attribute.
    *
    * @return   The current value of MAX-WIDTH-CHARS attribute.
    */
   public decimal getMaxWidthChars();

   /**
    * Gets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @return   The current value of the KEEP-FRAME-Z-ORDER attribute.
    */
   public logical isKeepFrameZOrder();
   
   /**
    * Check if window has KEEP-Z-ORDER option set.
    * 
    * @return <code>true</code> if window has KEEP-Z-ORDER option set.
    */
   public logical isKeepZOrder();

   /**
    * Check if window has message area.
    * 
    * @return <code>true</code> if window has message area.
    */
   public logical isMessageArea();

   /**
    * Check if window size can be changed.
    * 
    * @return  <code>true</code> if window size can be changed.
    */
   public logical isResize();

   /**
    * Check if window has scrollbars.
    * 
    * @return  <code>true</code> if window has scrollbars.
    */
   public logical isScrollBars();

   /**
    * Check if window has message area.
    * 
    * @return  <code>true</code> if window has message area.
    */
   public logical isStatusArea();

   /**
    * Load icon for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   public logical loadIcon(character searchPath);

   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   public void setKeepFrameZOrder(boolean keep);

   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   public void setKeepFrameZOrder(logical keep);

   /**
    * Enable/disable KEEP-Z-ORDER option.
    * 
    * @param   logical
    *          New flag value.
    */
   public void setKeepZOrder(logical logical);

   /**
    * Set maximal height.
    * 
    * @param   decimal
    *          New option value.
    */
   public void setMaxHeight(NumberType decimal);

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   public void setMaxHeightChars(decimal max);

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   public void setMaxHeightChars(double max);

   /**
    * Set maximal width.
    * 
    * @param   decimal
    *          New option value.
    */
   public void setMaxWidth(NumberType decimal);

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   public void setMaxWidthChars(decimal max);

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   public void setMaxWidthChars(double max);

   /**
    * Enable/disable displaying of message area.
    * 
    * @param   logical
    *          New flag value.
    */
   public void setMessageArea(logical logical);

   /**
    * Enable/disable window resizing.
    * 
    * @param   logical
    *          New option value.
    */
   public void setResize(logical logical);

   /**
    * Enable/disable displaying of scrollbars.
    * 
    * @param   logical
    *          New option value.
    */
   public void setScrollBars(logical logical);

   /**
    * Enable/disable displaying of status area.
    * 
    * @param   logical
    *          New option value.
    */
   public void setStatusArea(logical logical);
}
