/*
** Module   : SourceNameMapper.java
** Abstract : Progress procedure file name to Java class name run-time 
**            mapping (bidirectional) 
**
** Copyright (c) 2004-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20050817   @22156 Created initial version.
** 002 GES 20060124   @24072 Rewrite to move to a directory based 
**                           approach for storage of the mapping data.
**                           In addition, the mapping data used at 
**                           runtime is stored in context-local instances 
**                           to ensure that multiple configurations can
**                           be run simultaneously. Added a reverse
**                           mapping from Java to Progress names.
** 003 GES 20060201   @24185 Added external init option so that the class
**                           can be initialized from a source other than
**                           the directory.
** 004 NVS 20070126   @32009 "class-mapping" node of the directory is 
**                           searched as an account level standard search
**                           item.
** 005 ECF 20070220   @32186 Made map access thread-safe. Synchronized
**                           initialization of mapping data with read and
**                           write requests. This fixes a race condition
**                           which was resulting in an NPE.
** 006 GES 20080722   @39167 Added a fallback for convert() to allow a
**                           parameter which is already an existing
**                           fully qualified class name, to be returned.
**                           Major rewrite to load the name mapping data
**                           from a JAR file resource.  The new resource
**                           has a much simpler format.
** 007 SVL 20090310   @41448 Keys into methodMap for the Progress
**                           proc/function name to Java name mapping
**                           are always lower-cased.
** 008 GES 20090515   @42235 Import change.
** 009 ECF 20101004          Handle unqualified file name references in
**                           convertName method.
** 010 CA  20121220          Reworked the mapping to include the new emitted
**                           info (by collect_names.rules), needed by the
**                           persistent and super procedures implementation.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import java.net.*;
import org.w3c.dom.*;

import com.goldencode.util.*;

/**
 * This class performs a bidirectional run-time mapping of a Progress
 * procedure file name to and from a Java class name.  It also supports
 * run-time translation of Progress internal procedure names into Java
 * method names.
 * <p>
 * <b>The mapping data is loaded from the directory into data structures that
 * are <code>static</code>.  This means that at this time this mapping data
 * is global to the entire server rather than being context-local.  The
 * reason for this is to reduce the amount of memory used for this data
 * since 2+ map entries exist for each original source file (thus these maps
 * each contain thousands of entries for any reasonably sized
 * application).</b>
 * <p>
 * <b>The current implementation assumes a single root path for all Progress
 * procedure file names and single base package path for the generated Java
 * classes.  Despite these assumptions it is possible to translate
 * Progress procedure names incorrectly (to a Java class name which does not
 * belong to the subtree specified by the single root or which does not 
 * belong to the base package path).  These problems are caused by using
 * mismatching absolute paths for Progress names or names which start with
 * a dot ('&#46;') for Java class names.</b> The leading dot is stripped
 * from the returned class name and is not returned as a result of a
 * mapping.
 * <p>
 * The source file name can be in one of the many forms: relative path,
 * absolute path, can contain references to current and to upper level
 * directories. Also, the source file name can use the file naming format of
 * the original operating system used to run the Progress program rather than
 * the current operating system used to run the converted Java application.
 * To this the directory contains values that define the original source
 * system's parameters such as path delimiter, file name delimiter and case 
 * sensitivity.
 */
public class SourceNameMapper
{
   /** Identify a native procedure. */
   private static final String TYPE_DLL_ENTRY = "DLL-ENTRY";

   /** Identify an internal procedure. */
   private static final String TYPE_PROCEDURE = "PROCEDURE";

   /** Identify an user defined function. */
   private static final String TYPE_FUNCTION  = "FUNCTION";
   
   /** Class mappings element. */
   private static final String ELEM_CMAP = "class-mapping";
   
   /** Methods mappings element. */
   private static final String ELEM_MMAP = "method-mapping";
   
   /** Parameter info element. */
   private static final String ELEM_PARAM = "parameter";
   
   /** Progress name attribute. */
   private static final String ATTR_PNAME = "pname";
   
   /** Java name attribute. */
   private static final String ATTR_JNAME = "jname";
   
   /** TYPE attribute. */
   private static final String ATTR_TYPE = "type";
   
   /** DLL-ENTRY attribute. */
   private static final String ATTR_DLL_ENTRY = "dll-entry";
   
   /** Parameter MODE attribute. */
   private static final String ATTR_PARAM_MODE = "mode";
   
   /** Progress procedure file name to Java class name map table. */
   private static Map<String, ExternalProgram> p2jMap = null;
   
   /** Progress Java class name to procedure file name map table. */
   private static Map<String, ExternalProgram> j2pMap = null;

   /** List of the directories in the PROPATH of the legacy system. */
   private static String[] propath = null; 
   
   /** Base path for all Java packages. */
   private static String pkgroot = null;
   
   /** Base path for Progress 4GL code. */
   private static String proroot = null;
   
   /** Name element separator used for the Progress file name data. */
   private static String fileSep = null; 
   
   /** Path separator used for the Progress file name data. */
   private static String pathSep = null;

   /** Case sensitivity flag for the Progress file name data. */
   private static boolean caseSens; 
   
   /**
    * Get the legacy 4gL name for the Java class name. 
    * 
    * @param   jname
    *          The external-program's converted java name.
    * 
    * @return  The legacy program name or <code>null</code> if the given java
    *          program is not registered in name_map.xml.
    */
   public static String convertJavaProg(String jname)
   {
      initMappingData();
      ExternalProgram extp = j2pMap.get(removePkg(jname));
      return extp == null ? null : extp.pname;
   }
   
   /**
    * Convert a Progress file name into Java class name. The input name can
    * be an absolute file name, a file name relative to current location or
    * a file name relative to some directory listed in the PROPATH
    * variable. Before processing, the input file name is processed by 
    * {@link #canonicalize}.
    * <p>
    * A simple lookup is tried first.  If this mapping does not exist then
    * each path segment of the PROPATH is tried in order.  The given program
    * name is check to see if it is prefixed with the PROPATH segment, if
    * so that prefix is removed and a lookup is done with this name. The
    * first match found ends the search.
    * <p> 
    * Absolute paths are converted into relative paths based on the existence
    * of a matching path in the legacy PROPATH (see 
    * {@link EnvironmentOps#getLegacySearchPath}).  To handle additional
    * absolute paths, one must add the appropriate path prefix to the legacy
    * PROPATH and it will be removed here.
    * <p>
    * This same PROPATH processing is how names relative to the current 
    * directory can be handled.  This is the only way to handle that case
    * since the current directory in the source system is runtime information
    * that is not available here.
    * <p>
    * The legacy PROPATH is the key configuration tool by which one specifies
    * to this method how to remove absolute and current directory relative
    * prefixes.  This means that the legacy PROPATH that is used here may
    * need to be different than the PROPATH on the original system.
    * <p>
    * As a fallback, if no mapping can be found and if the given
    * <code>progName</code> is a fully qualified Java class name which
    * references a valid Java class, then the <code>progName</code> is
    * returned unchanged.
    *
    * @param    progName
    *           Progress file name.
    * 
    * @return   Java class name for the given Progress file name or
    *           <code>null</code> if no appropriate mapping is present.
    */
   public static String convertName(String progName)
   {
      initMappingData();
      
      // check the simple form of the name
      ExternalProgram extProg = p2jMap.get(canonicalize(progName));
      
      // if we didn't find a match AND if there is a propath, then
      // remove any paths that are in the propath from the front, this is 
      // how we remove an absolute path name, then try a lookup
      if (extProg == null && propath != null)
      {  
         String possible = null;
         
         for (int i = 0; i < propath.length; i++)
         {
            if (progName.startsWith(propath[i]))
            {
               // remove the prefix
               possible = progName.substring(propath[i].length());
               
               // remove any extra file separator
               if (possible.startsWith(fileSep))
               {
                  possible = possible.substring(1);
               }
               
               // check for a match
               extProg = p2jMap.get(canonicalize(possible));

               if (extProg != null)
               {
                  // first match exits the loop
                  break;
               }
            }
         }
         
         // if we still didn't find a match, strip the canonicalized base path
         // off each canonicalzed component of the propath, then prepend that
         // relative path to the name we have and try to look it up
         if (extProg == null)
         {
            try
            {
               String base = Utils.canonicalizePath(proroot,
                                                    File.separator,
                                                    fileSep);
               
               // make sure there is a base path (it only exists at conversion)
               if (base.length() > 0)
               {
                  for (int i = 0; i < propath.length; i++)
                  {
                     String next = Utils.canonicalizePath(propath[i],
                                                          File.separator,
                                                          fileSep);
                     
                     if (!next.startsWith(base))
                     {
                        continue;
                     }
                     
                     next = next.substring(base.length());
                     StringBuilder buf = new StringBuilder(next);
                     buf.append(fileSep);
                     buf.append(progName);
                     possible = canonicalize(buf.toString());
                     
                     extProg = p2jMap.get(canonicalize(possible));
                     
                     if (extProg != null)
                     {
                        break;
                     }
                  }
               }
            }
            catch (IOException exc)
            {
               // do nothing;  can't go further
            }
         }
      }
      
      // fallback processing
      if (extProg == null)
      {
         // enable support for fully qualified Java class names if they are
         // passed in directly
         try
         {
            Class.forName(progName);
            
            // if we are here, then the progName must have been a fully
            // qualified Java class name to a class that exists so we return
            // it unchanged
            return progName;
         }
         
         catch (ClassNotFoundException cnfe)
         {
            // do nothing, the given program can't be found as a Java class
         }
      }
      
      return extProg == null ? null : addPkg(extProg.jname);
   }
 
   /**
    * Get the Progress source file name (in relative filename form) using
    * the Java class name (which can be fully qualified).  If the method
    * name given is not <code>null</code>, then this method will additionally
    * lookup the internal procedure or function name represented by the
    * method and the two values will be returned as a single string of the
    * form "internal_proc_or_func_name external_source_file_name".
    * <p>
    * The internal procedure/function name lookup is best efforts.  If one
    * cannot be properly resolved, but the external procedure is found, then
    * the external procedure name will still be returned (by itself).
    * 
    * @param    javaName
    *           The Java class name to map to a Progress source file name.
    *           May not be <code>null</code>.
    * @param    methName
    *           The Java method name to map to an internal procedure or
    *           function name.  May be <code>null</code>.
    * 
    * @return   The relative Progress source file name name or
    *           <code>null</code> if no such mapping exists.
    */
   public static String convertStackEntry(String javaName, String methName)
   {
      initMappingData();
      
      javaName = removePkg(javaName);

      ExternalProgram extProg = j2pMap.get(javaName);
      String result = extProg.pname;
      
      // try to lookup the internal procedure or function name
      if (methName != null)
      {
         // get the map of methods for this class
         InternalEntry procName = extProg.searchLocalj(methName);
         if (procName != null)
         {
            StringBuffer sb = new StringBuffer(procName.pname);
            sb.append(" ").append(extProg.pname);
            result = sb.toString();
         }
      }
      
      return result;
   }
   
   /**
    * This method loads all mapping data from an external source. DO NOT
    * USE THIS UNLESS YOU KNOW EXACTLY WHAT YOU ARE DOING!!!!!  All data
    * structures are overwritten (including anything read from the
    * directory).  No safety or integrity checking is done at all.
    *
    * @param    p2jMap
    *           Progress procedure file name to Java class name map table.
    * @param    j2pMap
    *           Progress Java class name to procedure file name map table.
    * @param    propath
    *           List of the directories in the PROPATH of the legacy system.
    * @param    pkgroot
    *           Base path for all Java packages.
    * @param    proroot
    *           Base path for 4GL code.
    * @param    fileSep
    *           Name element separator used for the Progress file name data.
    * @param    pathSep
    *           Path separator used for the Progress file name data.
    * @param    caseSens
    *           Case sensitivity flag for the Progress file name data.
    */
   public synchronized static void overwriteMappingData(Map      p2jMap,
                                                        Map      j2pMap,
                                                        String[] propath,
                                                        String   pkgroot,
                                                        String   proroot,
                                                        String   fileSep,
                                                        String   pathSep,
                                                        boolean  caseSens)
   {
      SourceNameMapper.p2jMap    = p2jMap;
      SourceNameMapper.j2pMap    = j2pMap;
      SourceNameMapper.propath   = propath; 
      SourceNameMapper.pkgroot   = pkgroot;
      SourceNameMapper.proroot   = proroot;
      SourceNameMapper.fileSep   = fileSep; 
      SourceNameMapper.pathSep   = pathSep;
      SourceNameMapper.caseSens  = caseSens;
   }

   /**
    * Get the java name of the passed internal-entry and 4GL program.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    * @param    function
    *           <code>true</code> if the search is for a user-defined 
    *           function.
    *           
    * @return   The Java name of the given internal-entry, if this entry 
    *           exists in the given program. Else, <code>null</code>
    */
   public static String getMethodName(String  pname, 
                                      String  ieName, 
                                      boolean function)
   {
      initMappingData();
      
      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return null;
      }
      
      InternalEntry method = extProg.searchp(ieName, function);
      if (method == null)
      {
         return null;
      }
      
      return method.jname;
   }

   /**
    * Get the dll-entry attribute of the passed internal-entry.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    *           
    * @return   The dll-entry attribute of the passed internal-entry name,
    *           assuming the program exists and the internal-entry is a native
    *           procedure definition. Else, <code>null</code>
    */
   public static String getDll(String pname, String ieName)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return null;
      }
      
      InternalEntry ie = extProg.search(ieName, false);
      if (ie == null)
      {
         return null;
      }
      
      String type = ie.getAttribute(ATTR_TYPE);
      if (TYPE_DLL_ENTRY.equalsIgnoreCase(type))
      {
         return ie.getAttribute(ATTR_DLL_ENTRY);
      }
      
      return null;
   }

   /**
    * Get the string representation of the parameter modes for this external
    * procedure.  If there are no parameters, return null.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    *           
    * @return   See above.
    */
   public static String getParameterModes(String pname)
   {
      return getParameterModes(pname, null, false);
   }

   /**
    * Get the string representation of the parameter modes for this internal
    * entry. If there are no parameters, return null.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name. If null or the empty string, the
    *           external procedure's parameters are used.
    * @param    function
    *           <code>true</code> if this is a function call.
    *           
    * @return   See above.
    */
   public static String getParameterModes(String pname,
                                          String ieName,
                                          boolean function)
   {
      initMappingData();
      
      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return null;
      }

      if (ieName == null || "".equals(ieName))
      {
         return extProg.main.getParameterModes();
      }
      
      InternalEntry ie = extProg.search(ieName, function);
      
      return ie == null ? null : ie.getParameterModes();
   }
   
   /**
    * Get the 4GL-style signature of the given internal-entry.  If the program
    * does not exist or the internal-entry does not belong to this program,
    * return <code>null</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    *           
    * @return   See above.
    */
   public static String getSignature(String pname, String ieName)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return "";
      }
      
      if (ieName == null || "".equals(ieName))
      {
         return extProg.main.getSignature();
      }
      
      InternalEntry ief = extProg.search(ieName, true);

      // if both func and proc are with the same name, then func has priority
      InternalEntry ie = ief != null ? ief : extProg.search(ieName, false);
      
      return ie == null ? "" : ie.getSignature();
   }

   /**
    * Get the internal-entries attribute of the given external procedure.
    * If the program does not exist, return <code>null</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    prv
    *           Flag to include or not the private members.
    *           
    * @return   See above.
    */
   public static String getInternalEntries(String pname, boolean prv)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return null;
      }
      
      return extProg.getInternalEntries(prv);
   }

   /**
    * Check if the specified internal-entry is defined IN SUPER. If the program
    * does not exist or the internal entry does not belong to the program,
    * return <code>false</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    * @param    function
    *           Flag indicating if this is a function or a procedure.
    *
    * @return   See above.
    */
   public static boolean isSuper(String pname, 
                                 String ieName,
                                 boolean function)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return false;
      }
      
      InternalEntry ie = extProg.searchp(ieName, function);
      if (ie == null)
      {
         return false;
      }
      
      return ie.isSuper();
   }

   /**
    * Check if the specified internal-entry is a function. If the program
    * does not exist or the internal entry does not belong to the program,
    * return <code>false</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    *
    * @return   See above.
    */
   public static boolean isFunction(String pname, String ieName)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return false;
      }
      
      InternalEntry ie = extProg.searchp(ieName, true);

      return ie != null;
   }

   /**
    * Check if the specified internal-entry is defined PRIVATE. If the program
    * does not exist or the internal entry does not belong to the program,
    * return <code>false</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    * @param    function
    *           Flag indicating if this is a function or a procedure.
    *
    * @return   See above.
    */
   public static boolean isPrivate(String pname,
                                   String ieName,
                                   boolean function)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return false;
      }
      
      InternalEntry ie = extProg.searchp(ieName, function);
      if (ie == null)
      {
         return false;
      }
      
      return ie.isPrivate();
   }

   /**
    * Check if the specified internal-entry is defined IN handle. If the 
    * program does not exist or the internal entry does not belong to the
    * program, return <code>false</code>.
    * 
    * @param    pname
    *           Legacy 4GL program name.
    * @param    ieName
    *           Legacy internal-entry name.
    * @param    function
    *           Flag indicating if this is a function or a procedure.
    *
    * @return   See above.
    */
   public static boolean isInHandle(String pname,
                                    String ieName,
                                    boolean function)
   {
      initMappingData();

      ExternalProgram extProg = p2jMap.get(canonicalize(pname));
      if (extProg == null)
      {
         return false;
      }
      
      InternalEntry ie = extProg.searchp(ieName, function);
      if (ie == null)
      {
         return false;
      }
      
      return ie.isInHandle();
   }

   /**
    * This method loads all mapping data from a URL which is calculated
    * using the <code>pkgroot</code> and the <code>name_map.xml</code> with
    * all "." characters converted to file separators and an extra file
    * separator added before the file name portion. This URL is expected to
    * be found in the application's JAR file. Note that this will silently
    * return if the mapping data is not found.  Any file system configuration
    * data will be read from the directory via the {@link EnvironmentOps}
    * helper.
    */
   private synchronized static void initMappingData()
   {
      if (p2jMap != null)
      {
         return;
      }
      
      try
      {
         p2jMap    = new HashMap<String, ExternalProgram>();
         j2pMap    = new HashMap<String, ExternalProgram>();
         
         // lookup the descriptors for the original file system
         fileSep   = EnvironmentOps.getLegacyFileSeparator();
         pathSep   = EnvironmentOps.getLegacyPathSeparator();
         caseSens  = EnvironmentOps.getLegacyCaseSensitive();
         pkgroot   = Utils.getDirectoryNodeString(null,
                                                  "file-system/pkgroot",
                                                  "");
         
         // this is only a conversion-time configuration value ("basepath")
         proroot   = "";
         
         // get the original propath and split it into pieces
         String searchpath = EnvironmentOps.getLegacySearchPath();
         
         if (!caseSens)
         {
            searchpath = searchpath.toLowerCase();
         }
         
         propath = searchpath.split(pathSep);
         
         // build the fully qualified name of the name map XML file
         StringBuilder sb = new StringBuilder();
         sb.append(pkgroot.replaceAll("\\.", fileSep));
         
         if (!fileSep.equals(sb.substring(sb.length() - 1)))
         {
            sb.append(fileSep);
         }
         
         sb.append("name_map.xml");
         
         // create a URL to allow the name map to be loaded as a resource
         URL url = ClassLoader.getSystemResource(sb.toString());
         
         // safety first
         if (url == null)
         {
            // TODO: log this
            // no mapping data exists
            return;
         }
         
         Document dom = XmlHelper.parse(url.openStream());
         
         // find the root node
         Element root = dom.getDocumentElement();
         
         if (root == null)
         {
            // TODO: log this
            // no data in the resource
            return;
         }
         
         // enumerate all child nodes (there will be one for each source
         // file that can be mapped)
         NodeList cmaps = root.getElementsByTagName(ELEM_CMAP);
         
         if (cmaps == null)
         {
            // TODO: log this
            // no class mappings
            return;
         }
         
         int len = cmaps.getLength();
         
         for (int i = 0; i < len; i++)
         {
            Element classMap = (Element) cmaps.item(i);
            
            String pname = classMap.getAttribute(ATTR_PNAME);
            String jname = classMap.getAttribute(ATTR_JNAME);
            
            if (pname == null || pname.length() == 0 ||
                jname == null || jname.length() == 0)
            {
               String errmsg = "Malformed class mapping entry for " +
                               classMap.toString();
               throw new IllegalStateException(errmsg);
            }
            
            // remove case if required
            if (!caseSens)
            {
               pname = pname.toLowerCase();
            }
                                                    
            // store bidirectional mappings for the progress relative
            // filename to/from the java relative file name mapping
            ExternalProgram ep = buildExternalProgram(pname, jname, classMap);
            p2jMap.put(pname, ep);
            j2pMap.put(jname, ep);
         }
      }
   
      catch (Exception e)
      {
         e.printStackTrace();
         throw new RuntimeException("Unable to load name mapping data.");
      }
   }

   /**
    * Collect the {@link ExternalProgram} info contained in the <code>classMap
    * </code> node.
    * 
    * @param    pname
    *           The legacy 4GL program name.
    * @param    jname
    *           The converted class name.
    * @param    classMap
    *           The DOM node containing the info for this external program.
    *           
    * @return   An {@link ExternalProgram} instance for this DOM node.
    */
   private static ExternalProgram buildExternalProgram(String pname, 
                                                       String jname, 
                                                       Element classMap)
   {
      ExternalProgram extProg = new ExternalProgram(pname, jname);
      
      // enumerate method mapping nodes
      NodeList mmaps = classMap.getElementsByTagName(ELEM_MMAP);
      
      if (mmaps == null)
      {
         return extProg;
      }
      
      int mlen = mmaps.getLength();
      
      if (mlen > 0)
      {
         extProg.ieList = new ArrayDeque<InternalEntry>();
      }
      
      Set<String> defFuncs = new HashSet<String>();
      Set<String> forwardFuncs = new HashSet<String>();
      
      // process each method mapping
      // 4GL allows same name for a function or procedure, as long as they
      // are not internal. they appear in the internal-entry list, but the
      // get-signature call will see only the function (for this "feature" to
      // work, the procedure must be defined before the function, else 4GL
      // gets compile error).
      for (int j = 0; j < mlen; j++)
      {
         Element methMap = (Element) mmaps.item(j);
         
         String iepname = methMap.getAttribute(ATTR_PNAME);
         String iejname = methMap.getAttribute(ATTR_JNAME);
         
         if (iepname == null || iepname.length() == 0 ||
             iejname == null || iejname.length() == 0)
         {
            String errmsg = "Malformed method mapping entry for " +
                            methMap.toString();
            throw new IllegalStateException(errmsg);
         }

         InternalEntry ie = buildMethod(iepname, iejname, methMap);
         
         // the FORWARD functions are emitted in the name_map.xml only to
         // preserve the order.
         if (ie.isFunction())
         {
            String forward = methMap.getAttribute("forward");
            if ("true".equalsIgnoreCase(forward))
            {
               forwardFuncs.add(iepname);
            }
            else
            {
               defFuncs.add(iepname);

               // if the function already exists as forward, take the parameter
               // names from here and update the forward declaration
               if (forwardFuncs.contains(iepname))
               {
                  if (ie.parameters.size() > 0)
                  {
                     InternalEntry ief = extProg.search(iepname, true);
                     ief.parameters = ie.parameters;
                  }

                  continue;
               }
            }
         }
         
         // store the bidirectional Progress relative
         // proc/function name from/to Java method name
         // mapping
         extProg.ieList.add(ie);
         
         Map<String, InternalEntry> p2j = null;
         Map<String, InternalEntry> j2p = null;
         if (ie.isFunction())
         {
            if (extProg.p2jf.isEmpty())
            {
               extProg.p2jf = new HashMap<String, InternalEntry>();
               extProg.j2pf = new HashMap<String, InternalEntry>();
            }
            p2j = extProg.p2jf;
            j2p = extProg.j2pf;
         }
         else
         {
            if (extProg.p2jp.isEmpty())
            {
               extProg.p2jp = new HashMap<String, InternalEntry>();
               extProg.j2pp = new HashMap<String, InternalEntry>();
            }
            p2j = extProg.p2jp;
            j2p = extProg.j2pp;
         }
         p2j.put(iepname.toLowerCase(), ie);
         j2p.put(iejname, ie);
      }
      
      // remove all functions defined as forward but without body
      forwardFuncs.removeAll(defFuncs);
      for (String iepname : forwardFuncs)
      {
         InternalEntry ie = extProg.search(iepname, true);
         extProg.ieList.remove(ie);
      }
      
      extProg.main = new InternalEntry("", "");
      extProg.main.attributes.put("type", "MAIN");
      buildParameters(classMap, extProg.main);
      
      return extProg;
   }
   
   /**
    * Collect the {@link InternalEntry} info contained in the <code>methMap
    * </code> node.
    * 
    * @param    pname
    *           The legacy 4GL internal-entry name.
    * @param    jname
    *           The converted internal-entry name.
    * @param    methMap
    *           The DOM node containing the info for this internal-entry.
    *           
    * @return   An {@link ExternalProgram} instance for this DOM node.
    */
   private static InternalEntry buildMethod(String pname, 
                                            String jname, 
                                            Element methMap)
   {
      InternalEntry m = new InternalEntry(pname, jname);
      
      buildParameters(methMap, m);
      
      // add the method's attributes
      int attrLen = methMap.getAttributes().getLength();
      for (int i = 0; i < attrLen; i++)
      {
         Node n = methMap.getAttributes().item(i);
         String attrName = n.getNodeName().toLowerCase();
         String attrValue = n.getNodeValue();

         m.attributes.put(attrName, attrValue);
      }

      return m;
   }
   
   /**
    * Build the parameters for the specified internal entry.
    * 
    * @param   node
    *          The current node from where the parameters are read.
    * @param   ie
    *          The internal entry where to save the parameters.
    */
   private static void buildParameters(Element node, InternalEntry ie)
   {
      // this enumerates all parameter nodes under the node instance (on any
      // level)
      NodeList params = node.getElementsByTagName(ELEM_PARAM);
      
      if (params == null)
      {
         return;
      }
      
      int plen = params.getLength();
      
      if (plen > 0)
      {
         ie.parameters = new ArrayList<Parameter>();
      }
      
      // process each method mapping
      for (int j = 0; j < plen; j++)
      {
         Element param = (Element) params.item(j);
         // keep only parameters which belong to the given node
         if (param.getParentNode() != node)
         {
            continue;
         }
         
         String ppname = param.getAttribute(ATTR_PNAME);
         String pjname = param.getAttribute(ATTR_JNAME);
         String ptype  = param.getAttribute(ATTR_TYPE);
         String pmode  = param.getAttribute(ATTR_PARAM_MODE);
         
         if (!("TABLE".equalsIgnoreCase(ptype) || 
               "BUFFER".equalsIgnoreCase(ptype)) &&
             (ppname == null || ppname.length() == 0 ||
              pjname == null || pjname.length() == 0))
         {
            String errmsg = "Malformed parameter entry for " +
                            param.toString() + ", in method " + 
                            node.toString();
            throw new IllegalStateException(errmsg);
         }

         Parameter p = new Parameter(ppname, pjname, ptype, pmode);

         // add the parameter's attributes
         int attrLen = param.getAttributes().getLength();
         for (int i = 0; i < attrLen; i++)
         {
            Node n = param.getAttributes().item(i);
            String attrName = n.getNodeName().toLowerCase();
            String attrValue = n.getNodeValue();

            p.attributes.put(attrName, attrValue);
         }

         ie.parameters.add(p);
      }
   }
   
   /**
    * Build the 4GL-style signature of the given {@link InternalEntry} 
    * instance.
    * 
    * @param    internalEntry
    *           The {@link InternalEntry} instance for which the signature is
    *           needed.
    *           
    * @return   The internal-entry's signature.
    */
   private static String buildSignature(InternalEntry internalEntry)
   {
      // The format is: TYPE,return type,param1,param2,param3 where:
      // - TYPE is either PROCEDURE, FUNCTION, EXTERN or DLL-ENTRY
      // - return type is the 4GL name of the return data type
      // - param# is one of: 
      //     INPUT varname TYPE
      //     OUTPUT varname TYPE or INPUT-OUTPUT varname TYPE.
      //   Varname is either the original 4GL variable name or "arg#", if this
      //   is a header definition (i.e. an IN SUPER or IN handle) which had
      //   its parameter set as:
      //   "function f1 returns int (input int) in super."
      String type = internalEntry.getAttribute("type");
      String returns = internalEntry.getAttribute("returns");
      if (returns == null)
      {
         returns = "";
      }
      if (internalEntry.isSuper() || internalEntry.isInHandle())
      {
         returns = returns.toUpperCase();
      }
      else
      {
         returns = returns.toLowerCase();
      }
      
      StringBuilder sb = new StringBuilder();
      if (internalEntry.isInHandle())
      {
         type = "EXTERN";
      }
      sb.append(type).append(",");
      sb.append(returns);
      if (internalEntry.parameters.size() == 0)
      {
         sb.append(",");
      }
      else
      {
         int idx = 0;
         for (Parameter p : internalEntry.parameters)
         {
            idx++;
            
            sb.append(",");
            if (p.mode.length() > 0)
            {
               sb.append(p.mode).append(" ");
            }

            if ("BUFFER".equalsIgnoreCase(p.type))
            {
               String buffer = p.getAttribute("buffer");
               String forTable = p.getAttribute("for");

               sb.append("BUFFER").append(" ");
               sb.append(buffer).append(" ");
               sb.append(forTable);
            }
            else if ("TABLE".equalsIgnoreCase(p.type))
            {

               String table = p.getAttribute("table");
               String append = p.getAttribute("append");

               sb.append("TABLE").append(" ");
               sb.append(table);
               if ("true".equalsIgnoreCase(append))
               {
                  sb.append(" ").append("APPEND");
               }
            }
            else
            {
               String pname = p.pname;
               if (pname == null || pname.length() == 0)
               {
                  pname = String.format("arg%d", idx);
               }

               sb.append(pname).append(" ");
               sb.append(p.type);
            }
         }
      }

      return sb.toString();
   }

   /**
    * Build the internal-entries 4GL attribute for the given external program.
    * 
    * @param    extProg
    *           The {@link ExternalProgram} instance.
    * @param    prv
    *           Flag to include or not the private members.
    *           
    * @return   See above.
    */
   private static String buildInternalEntries(ExternalProgram extProg, 
                                              boolean prv)
   {
      // - PRIVATE members are not visible
      // - prototypes (such as IN SUPER or IN handle functions/procs) appear 
      //   in this list 
      // - FORWARD functions without implementations do not appear
      // - first are procedures, then the functions.  IN SUPER procedures 
      // are put last (in reverse order).

      StringBuilder sbSuperProcs = new StringBuilder();
      StringBuilder sbProcs = new StringBuilder();
      StringBuilder sbFuncs = new StringBuilder();
      List<String> superProcs = new LinkedList<String>();
      
      for (InternalEntry ie : extProg.ieList)
      {
         StringBuilder sb = null;
         if (ie.isPrivate() && !prv)
         {
            continue;
         }

         if (ie.isFunction())
         {
            sb = sbFuncs;
         }
         else
         {
            if (ie.isSuper())
            {
               superProcs.add(0, ie.pname);
               continue;
            }

            sb = sbProcs;
         }
         
         if (sb.length() > 0)
         {
            sb.append(',');
         }
         
         sb.append(ie.pname);
      }

      for (String psuper : superProcs)
      {
         if (sbSuperProcs.length() > 0)
         {
            sbSuperProcs.append(',');
         }
         
         sbSuperProcs.append(psuper);
      }
      
      StringBuilder sb = new StringBuilder();
      sb.append(sbProcs);
      if (sb.length() > 0 && sbFuncs.length() > 0)
      {
         sb.append(',');
      }
      sb.append(sbFuncs);
      if (sb.length() > 0 && sbSuperProcs.length() > 0)
      {
         sb.append(',');
      }
      sb.append(sbSuperProcs);
      
      return sb.toString();
   }

   /**
    * Convert the given Progress file name into a simple relative path
    * form.  Any leading file separator is removed.  All references
    * to directories <code>&#46;</code> and <code>&#46;&#46;</code> are
    * processed.  Empty path elements are removed.
    * <p>
    * If {@link #caseSens} is false, the program name is lower-cased.
    * 
    * @param    progName
    *           Input name.
    * 
    * @return   The relative name.
    */
   private static String canonicalize(String progName)
   {
      // honor case-insensitivity
      if (!caseSens)
      {
         progName = progName.toLowerCase();
      }
      
      // backslash is regex escape character, so it needs to be escaped itself
      String splitOn = ("\\".equals(fileSep) ? "\\\\" : fileSep);
      String[] lst = progName.split(splitOn);
      
      LinkedList list = new LinkedList();
      
      for (int i = 0; i < lst.length; i++)
      {
         if (lst[i].equals("") || lst[i].equals("."))
            continue;
         
         if (lst[i].equals(".."))
         {
            if (list.size() > 0)
               list.removeLast();
            
            continue;
         }
         
         list.addLast(lst[i]);
      }
      
      StringBuffer buf = new StringBuffer(progName.length());
      
      for (Iterator iter = list.iterator(); iter.hasNext();)
      {
         String element = (String) iter.next();
         
         buf.append(element);
         
         if (iter.hasNext())
            buf.append(fileSep);
      }
      
      return buf.toString();
   }

   /**
    * Construct a full class name (including packages) for the stored package
    * root and the given class name.
    * 
    * @param    name
    *           Class name.
    * 
    * @return   The fully qualified class name.
    */
   private static String addPkg(String name)
   {
      initMappingData();

      if (name == null)
         return null;
      
      if (pkgroot == null)
         return name;
      
      StringBuffer buf = new StringBuffer(pkgroot);
      
      buf.append(".");
      buf.append(name);
      
      return buf.toString(); 
   }
   

   /**
    * Remove the package prefix from the given class name.
    * 
    * @param    name
    *           Class name.
    * 
    * @return   The class name with the root package prefix removed.
    */
   private static String removePkg(String name)
   {
      initMappingData();

      if (name == null)
         return null;
      
      if (pkgroot == null)
         return name;
      
      if (name.startsWith(pkgroot))
      {
         name = name.substring(pkgroot.length() + 1);
      }
      
      return name;
   }

   /**
    * Container for all external program information.
    */
   private static class ExternalProgram
   {
      /** Legacy 4GL name for this external program. */
      String pname = null;

      /** Converted Java name for this external program. */
      String jname = null;

      /** 
       * List of all internal entries, as they appear in the name_map file.
       * This is needed to provide consistency for the internal-entries 
       * attribute (as a function and a procedure may have the same name,
       * a map can not be used).
       */
      ArrayDeque<InternalEntry> ieList = new ArrayDeque<InternalEntry>();
      
      /**
       * Provide fast access to the java name of a specified legacy function
       * name.
       */
      Map<String, InternalEntry> p2jf = Collections.EMPTY_MAP;

      /**
       * Provide fast access to the java name of a specified legacy procedure
       * name.
       */
      Map<String, InternalEntry> p2jp = Collections.EMPTY_MAP;

      /**
       * Provide fast access to the legacy name of a specified java function
       * name.
       */
      Map<String, InternalEntry> j2pf = Collections.EMPTY_MAP;

      /**
       * Provide fast access to the legacy name of a specified java procedure
       * name.
       */
      Map<String, InternalEntry> j2pp = Collections.EMPTY_MAP;

      /** 
       * The <code>internal-entry</code> attribute for this external program,
       * without private members.
       */
      String internalEntries = null;

      /** 
       * The <code>internal-entry</code> attribute for this external program,
       * with private members.
       */
      String internalEntriesPrv = null;
      
      /**
       * The internal-entry associated with the MAIN (external) procedure.
       * .
       */
      InternalEntry main = null;

      /**
       * Basic c'tor.
       * 
       * @param    pname
       *           Legacy 4GL name for this external program.
       * @param    jname
       *           Converted Java name for this external program.
       */
      ExternalProgram(String pname, String jname) 
      {
         this.pname = pname;
         this.jname = jname;
      }
      
      /**
       * Gets the <code>internal-entry</code> attribute for this external
       * program.  If this is the first access, it initializes it.
       * 
       * @param    prv
       *           Flag to include or not the private members.
       *           
       * @return   See above.
       */
      String getInternalEntries(boolean prv)
      {
         if (prv)
         {
            if (internalEntriesPrv == null)
            {
               internalEntriesPrv = buildInternalEntries(this, prv);
            }
            
            return internalEntriesPrv;
         }
         else
         {
            if (internalEntries == null)
            {
               internalEntries = buildInternalEntries(this, prv);
            }
            
            return internalEntries;
         }
      }

      /**
       * Search the {@link #ieList internal-entry list} for the specified 
       * name, using the <code>function</code> flag to search only for 
       * functions or only for procedures.
       * 
       * @param    pname
       *           Legacy 4GL name for the search internal-entry.
       * @param    function
       *           <code>true</code> if search for a function
       * 
       * @return   The {@link InternalEntry} instance, or <code>null</code> if
       *           it was not found.
       */
      InternalEntry search(String pname, boolean function)
      {
         for (InternalEntry ie : ieList)
         {
            if (function == ie.function && ie.pname.equalsIgnoreCase(pname))
            {
               return ie;
            }
         }
         
         return null;
      }

      /**
       * Search the fast-access p2j maps for a function or procedure, using 
       * the <code>function</code> flag.
       * 
       * @param    pname
       *           Legacy 4GL name for the search internal-entry.
       * @param    function
       *           <code>true</code> if search for a function
       * 
       * @return   The {@link InternalEntry} instance, or <code>null</code> if
       *           it was not found.
       */
      InternalEntry searchp(String pname, boolean function)
      {
         Map<String, InternalEntry> p2j = function ? p2jf : p2jp;
         
         return p2j.get(pname.toLowerCase());
      }

      /**
       * Search the fast-access j2p maps for a function or procedure. This
       * ignores the IN SUPER and IN HANDLE entries, as it is expected to
       * always find a method with its body in this {@link ExternalProgram}.
       * 
       * @param    jname
       *           Java method name.
       * 
       * @return   The {@link InternalEntry} instance, or <code>null</code> if
       *           it was not found.
       */
      InternalEntry searchLocalj(String jname)
      {
         // in this case, we need only internal entries which body resides
         // in this procedure. so, IN SUPER and IN HANDLE entries are ignored.
         
         InternalEntry ie = j2pf.get(jname);
         if (ie != null && !ie.isInHandle() && !ie.isSuper())
         {
            return ie;
         }
         
         ie = j2pp.get(jname);
         if (ie != null && !ie.isInHandle() && !ie.isSuper())
         {
            return ie;
         }

         return null;
      }
   }
   
   /**
    * Container for legacy info belonging to an internal entry.
    */
   private static class InternalEntry
   {
      /** Legacy 4GL name for this internal-entry. */
      final String pname;

      /** Converted Java name for this internal-entry. */
      final String jname;
      
      /**
       * The signature of this internal entry, as returned by the 
       * <code>get-signature</code> 4GL API.
       */
      String signature = null;
      
      /**
       * The string representation of the parameter modes for this internal
       * entry. 
       */
      String parameterModes = null;
      
      /** The parameter list of this internal entry. */
      List<Parameter> parameters = Collections.EMPTY_LIST;

      /** A map containing all the internal entry's attributes. */
      Map<String, String> attributes = new HashMap<String, String>();
      
      /** 
       * Flag indicating if this is a function (when set to 
       * <code>true</code>). Gets populated on first {@link #isFunction()} call.
       */
      private Boolean function = null;

      /** 
       * Flag indicating if this is a IN SUPER internal entry (when set to 
       * <code>true</code>). Gets populated on first {@link #isSuper()} call.
       */
      private Boolean isSuper = null;

      /** 
       * Flag indicating if this is a IN handle internal entry (when set to 
       * <code>true</code>). Gets populated on first {@link #isInHandle()()} 
       * call.
       */
      private Boolean inHandle = null;

      /** 
       * Flag indicating if this is a PRIVATE internal entry (when set to 
       * <code>true</code>). Gets populated on first {@link #isPrivate()()} 
       * call.
       */
      private Boolean isPrivate = null;

      /**
       * Basic c'tor.
       * 
       * @param    pname
       *           Legacy 4GL name for this internal-entry.
       * @param    jname
       *           Converted Java name for this internal-entry.
       */
      InternalEntry(String pname, String jname)
      {
         this.pname = pname;
         this.jname = jname;
      }
      
      /**
       * Return <code>true</code> if this internal-entry has its 
       * <code>in-super</code> attribute set to <code>"true"</code>.
       * 
       * @return   See above.
       */
      boolean isSuper()
      {
         if (isSuper == null)
         {
            isSuper = "true".equalsIgnoreCase(getAttribute("in-super"));
         }
         
         return isSuper;
      }
      
      /**
       * Return <code>true</code> if this internal-entry has its 
       * <code>private</code> attribute set to <code>"true"</code>.
       * 
       * @return   See above.
       */
      boolean isPrivate()
      {
         if (isPrivate == null)
         {
            isPrivate = "true".equalsIgnoreCase(getAttribute("private"));
         }
         
         return isPrivate;
      }
      
      /**
       * Return <code>true</code> if internal-entry is a function (its 
       * {@link SourceNameMapper#ATTR_TYPE} is set 
       * to {@link SourceNameMapper#TYPE_FUNCTION}).
       * 
       * @return   See above.
       */
      boolean isFunction()
      {
         if (function == null)
         {
            function = TYPE_FUNCTION.equalsIgnoreCase(getAttribute(ATTR_TYPE));
         }
         
         return function;
      }

      /**
       * Return <code>true</code> if this internal-entry has its 
       * <code>in-handle</code> attribute set to <code>"true"</code>.
       * 
       * @return   See above.
       */
      boolean isInHandle()
      {
         if (inHandle == null)
         {
            inHandle = "true".equalsIgnoreCase(getAttribute("in-handle"));
         }

         return inHandle;
      }

      /**
       * Get the value of the given attribute.  If the attribute is null or
       * is not set, return <code>null</code>.
       * 
       * @param    attrName
       *           The attribute's name.
       *           
       * @return   See above.
       */
      String getAttribute(String attrName)
      {
         if (attrName == null)
         {
            return null;
         }
         
         attrName = attrName.toLowerCase();
         
         return attributes.get(attrName);
      }

      /**
       * Gets the signature of this internal-entry, as returned by the 
       * <code>get-signature</code> API of an persistent procedure.
       * If this is the first access, it initializes the 
       * {@link #signature} field, using the
       * {@link SourceNameMapper#buildSignature(InternalEntry)}.
       * 
       * @return   See above.
       */
      String getSignature()
      {
         if (signature == null)
         {
            signature = buildSignature(this);
         }
         
         return signature;
      }

      /**
       * Gets the string representation of the parameter modes for 
       * internal-entry. If this is the first access, it initializes the 
       * {@link #parameterModes} field.
       * 
       * @return   See above.
       */
      String getParameterModes()
      {
         if (parameterModes == null)
         {
            if (parameters.size() == 0)
            {
               return null;
            }
            else
            {
               StringBuilder sb = new StringBuilder();
               for (Parameter p : parameters)
               {
                  String mode = p.mode;
                  String type = p.type;
                  if ("BUFFER".equalsIgnoreCase(type))
                  {
                     mode = "B";
                  }
                  else if ("INPUT".equalsIgnoreCase(mode))
                  {
                     mode ="I";
                  }
                  else if ("OUTPUT".equalsIgnoreCase(mode))
                  {
                     mode ="O";
                  }
                  else if ("INPUT-OUTPUT".equalsIgnoreCase(mode))
                  {
                     mode ="U";
                  }
                  else
                  {
                     String errmsg = 
                        "Malformed parameter mode " + mode.toString() +
                        " for parameter " + pname;
                     throw new IllegalStateException(errmsg);
                  }
                  
                  sb.append(mode);
               }
               
               parameterModes = sb.toString();
            }
         }
         
         return parameterModes;
      }
   }
   
   /**
    * Container for all info related to an internal-entry's parameter.
    */
   private static class Parameter
   {
      /** A map containing all the parameter's attributes. */
      Map<String, String> attributes = new HashMap<String, String>();

      /** Legacy 4GL name for this parameter. */
      final String pname;

      /** Converted Java name for this parameter. */
      final String jname;

      /** The 4GL-style parameter type. */
      final String type;
      
      /** The mode of this parameter, one of INPUT, OUTPUT or INPUT-OUTPUT. */
      final String mode;

      /**
       * Get the value of the given attribute.  If the attribute is null or
       * is not set, return <code>null</code>.
       * 
       * @param    attrName
       *           The attribute's name.
       *           
       * @return   See above.
       */
      String getAttribute(String attrName)
      {
         if (attrName == null)
         {
            return null;
         }
         
         attrName = attrName.toLowerCase();
         
         return attributes.get(attrName);
      }

      /**
       * Basic c'tor.
       * 
       * @param    pname
       *           Legacy 4GL name for this parameter.
       * @param    jname
       *           Converted Java name for this parameter.
       * @param    type
       *           The 4GL-style parameter type.
       * @param    mode
       *           The mode of this parameter, one of INPUT, OUTPUT or 
       *           INPUT-OUTPUT.
       */
      Parameter(String pname, String jname, String type, String mode)
      {
         this.pname = pname;
         this.jname = jname;
         this.type = type;

         if (!"BUFFER".equalsIgnoreCase(type) &&
             !("INPUT".equalsIgnoreCase(mode)   || 
               "OUTPUT".equalsIgnoreCase(mode)  ||
               "INPUT-OUTPUT".equalsIgnoreCase(mode)))
         {
            String errmsg = "Malformed parameter mode " + mode.toString() +
                           " for parameter " + pname;
            throw new IllegalStateException(errmsg);
         }
         this.mode = mode.toUpperCase();
      }
   }
}
