/*
** Module   : raw.java
** Abstract : Progress 4GL compatible raw data type
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050906   @22563 Created initial version with support for a
**                           subset of raw processing.
** 002 ECF 20060403   @25321 Added copy constructor.
** 003 GES 20060724   @28173 Use forced assign in copy constructor.
** 004 GES 20090422   @41913 Converted to standard string formatting.
** 005 CA  20130123          Added c'tor which builds a new instance from a
**                           BaseDataType value, if possible. This should be
**                           used only when constructing an instance using the
**                           value returned by a DYNAMIC-FUNCTION call; 4GL
**                           does some special casting in this case, which is
**                           not implemented yet.
*/

package com.goldencode.p2j.util;

import org.apache.commons.codec.binary.Base64;

/**
 * A class that implements common services for a Progress 4GL compatible
 * raw data type whose data is mutable.  Only a subset of the Progress
 * language features are supported.
 * <p>
 * In addition to the implementation of many built-in functions
 * and operators, this class hides the representation of the 
 * <code>unknown value</code> and properly maintains the processing of
 * <code>unknown value</code>. 
 * <p>
 * Format string related processing is handled in this class, see method
 * {@link #toString(String)}.
 * <p>
 * All <code>RAW</code>features other than string formatting are implemented
 * in the common base class <code>BinaryData</code> which is shared with
 * <code>MEMPTR</code> types.
 *
 * @author    GES
 */
public class raw
extends BinaryData
{
   /** The maximum size of the binary data stored in an instance. */ 
   private static final int MAX_SIZE = 32 * 1024;
   
   /**
    * Default constructor, creates an instance that represents a 0 length
    * byte array.
    */
   public raw()
   {
      value = new byte[0];
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a DYNAMIC-FUNCTION call to the expected type (i.e
    * when passing that value as a parameter to another function or procedure
    * call).  In such cases, 4GL does some automatic conversion, which is not
    * implemented yet.
    * <p>
    * This version of the c'tor always sets the value to unknown.
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public raw(BaseDataType value)
   {
      this();
      // no-op
   }
   
   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * If the parameter's represents the <code>unknown value</code>, this
    * instance will also represent the <code>unknown value</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public raw(raw value)
   {
      assign(value, true);
   }
   
   /**
    * Constructor which creates an instance which is initialized based on
    * the given binary data.  The passed reference becomes the backing
    * reference for this instance (the data is not copied, only the
    * reference to the data is copied).
    *
    * @param    val
    *           The binary data to be used as the contents or if 
    *           <code>null</code> then the instance will represent the
    *           <code>unknown value</code>.
    */
   public raw(byte[] val)
   {
      super(val);
   }
   
   /**
    * Constructs an instance which converts the contents of the given
    * Progress 4GL compatible string form of a raw value into the proper
    * binary form.  When Progress converts a raw value into a string,  
    * it inserts 6 hexidecimal characters at the front of the string and
    * then formats all other data as a single Base64 encoded value (it
    * is NOT split into 76 character line segments).  The first 2 hexidecimal
    * characters are always <code>02</code> and the following 4 hexidcemial
    * characters are a hexidecimal representation of the length of the
    * resulting value.  These 6 preceding bytes are simply dropped.
    *
    * @param    value
    *           The Base64 value to be converted into the resulting
    *           instance binary data.
    */
   public raw(String value)
   {
      super(value);
   }
   
   /**
    * Constructs an instance which converts the contents of the given
    * Progress 4GL compatible string form of a raw value into the proper
    * binary form.  When Progress converts a raw value into a string,  
    * it inserts 6 hexidecimal characters at the front of the string and
    * then formats all other data as a single Base64 encoded value (it
    * is NOT split into 76 character line segments).  The first 2 hexidecimal
    * characters are always <code>02</code> and the following 4 hexidcemial
    * characters are a hexidecimal representation of the length of the
    * resulting value.  These 6 preceding bytes are simply dropped.
    *
    * @param    value
    *           The Base64 value to be converted into the resulting
    *           instance binary data.  If this is the 
    *           <code>unknown value</code>, then the result is the
    *           <code>unknown value</code>.
    */
   public raw(character value)
   {
      super(value);
   }
   
   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.
    */
   public BaseDataType duplicate()
   {
      raw r = new raw();
      
      // set the instance data
      r.assign(this, true);
      
      return r;
   }

   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      raw r = new raw();
      r.setUnknown();
      
      return r;
   }
   
   /**
    * Determines if the array of binary data stored in an instance of this
    * class must be limited in length.
    *
    * @return    <code>true</code> if the length must be limited.
    */
   public boolean isLengthLimited()
   {
      return true;
   }
   
   /**
    * Determines the maximum length of the array of binary data stored in an
    * instance of this class.
    *
    * @return    The length limit for binary data or -1 if this class has
    *            no limit.
    */
   public int lengthLimit()
   {
      return MAX_SIZE;
   }
   
   /**
    * Determines if the array of binary data stored in an instance of this
    * class must be automatically extended in length when operations
    * write past the current length.  Such extensions will pad with
    * <code>null</code> bytes.
    *
    * @return    <code>true</code> if the length must automatically extended.
    */
   public boolean isAutoExtend()
   {
      return true;
   }
   
   /**
    * Creates a defualt string representation of the instance data which
    * will be equivalent to the result of {@link #toStringExport}.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return   The formatted string.
    */
   public String toString()   
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data using the given
    * format string.  A <code>null</code> format string is equivalent to
    * the result of {@link #toStringExport}.  If the instance represents the
    * <code>unknown value</code>, a '?' will be returned.
    *
    * @param    fmt
    *           The format string to use.
    *
    * @return   The formatted string.
    */
   public String toString(String fmt)   
   {
      if (fmt == null)
      {
         return toStringExport();
      }
      
      if (unknown)
      {
         return "?";
      }
      
      if (fmt.length() == 0)
      {
         ErrorManager.recordOrThrowError(74,
                                         "Value  cannot be displayed using ");
                                         
         return "";
      }
      
      // use the normal character type to handle this data
      character tmp = new character(new String(value));
      
      return tmp.toString(fmt);
   }
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return   The 'message' formatted string.
    */
   public String toStringMessage()   
   {
      return toStringExport();
   }
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format in which the data is written as a base64 string prepended with 6
    * hexidecimal bytes of "header" information. Warning: this method is 
    * hard-coded to produce a header string starting with "02" (this seems
    * to be a fixed value for normal <code>raw</code> data in Progress, but
    * it may be different when records or other structured data is written)
    * and this is followed by 4 bytes specifying the length of the data.
    * <p>
    * If the instance represents the <code>unknown value</code>, a '?' will
    * be returned.
    * 
    * @return   The 'export' formatted string.
    */
   public String toStringExport()   
   {
      // quick out
      if (unknown)
      {
         return "?";
      }
      
      // always starts with 0x02
      StringBuffer sb = new StringBuffer("02");
      
      // add length as 4 hexidecimal chars
      sb.append(String.format("%04X", value.length));
      
      if (value.length > 0)
      {
         // add the contents of the array encoded base64
         String data = new String(Base64.encodeBase64(value));
         sb.append(data);
      }
      
      return sb.toString();
   }   
}
