/*
** Module   : rowid.java
** Abstract : Progress 4GL compatible rowid object
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20080721   @39290 Created initial version. Represents the
**                           Progress rowid data type. Much of the
**                           implementation was borrowed from the handle
**                           type.
** 002 ECF 20090212   @41302 Implemented a faster equals(). Handles most
**                           common cases without having to invoke the
**                           superclass' implementation.
** 003 CA  20130123          Added c'tor which builds a new instance from a
**                           BaseDataType value, if possible. This should be
**                           used only when constructing an instance using the
**                           value returned by a DYNAMIC-FUNCTION call; 4GL
**                           does some special casting in this case, which is
**                           not implemented yet.
*/

package com.goldencode.p2j.util;

import java.io.*;

/**
 * A class that represents a Progress 4GL compatible rowid data type.
 * Internally, the row identifier is stored as a 64-bit integer.
 */
public final class rowid
extends BaseDataType
{
   /** Stores the integral row identifier, unique within a database */
   private Long value = null;
   
   /**
    * Default constructor, creates an instance that represents the unknown
    * value.
    */
   public rowid()
   {
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a DYNAMIC-FUNCTION call to the expected type (i.e
    * when passing that value as a parameter to another function or procedure
    * call).  In such cases, 4GL does some automatic conversion, which is not
    * implemented yet.
    * <p>
    * This version of the c'tor always sets the value to unknown.
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public rowid(BaseDataType value)
   {
      this();
      // no-op
   }

   /**
    * Constructs an instance based upon an existing instance.
    *
    * @param   value
    *          The instance to copy.
    */
   public rowid(rowid value)
   {
      assign(value, false);
   }
   
   /**
    * Constructs an instance given a 64-bit integer value.
    * 
    * @param   value
    *          64-bit integer value;  should be unique across all records in a
    *          database.  May be <code>null</code> to indicate unknown value.
    */
   public rowid(Long value)
   {
      this.value = value;
   }
   
   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return  An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      return new rowid();
   }
   
   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return  A clone of this instance.  
    */
   public BaseDataType duplicate()
   {
      return (new rowid(this));
   }
   
   /**
    * An equality test which handles the most common cases with the least
    * amount of overhead.  The superclass implementation is invoked otherwise.
    * <p>
    * This implementation is consistent with {@link #hashCode()}.
    * 
    * @see com.goldencode.p2j.util.BaseDataType#equals(java.lang.Object)
    */
   public boolean equals(Object o)
   {
      if (this == o)
      {
         return true;
      }
      
      if (o instanceof rowid)
      {
         rowid that = (rowid) o;
         
         boolean u1 = (this.value == null);
         boolean u2 = (that.value == null);
         
         if (u1 && u2)
         {
            return true;
         }
         
         if (u1 || u2)
         {
            return false;
         }
         
         return (this.value.equals(that.value));
      }
      
      return super.equals(o);
   }
   
   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance. All instances of
    *          this class return the same hash code value, since all
    *          instances are inherently equal.
    */
   public int hashCode()
   {
      int result = 17;
      
      if (value != null)
      {
         result = 37 * result + value.hashCode();
      }
      
      return result;
   }
   
   /**
    * Returns the value of this instance as an <code>Long</code>.
    * <p>
    * If this instance represents the <code>unknown value</code>, a
    * <code>null</code> will be returned.
    *
    * @return  The value as a <code>Long</code> or <code>null</code> if
    *          this is <code>unknown</code>.
    */
   public Long getValue()
   {
      return value;
   }   
   
   /**
    * Determines if this instance represents the <code>unknown value</code>.
    *
    * @return  <code>true</code> if this instance is set to the
    *          <code>unknown value</code>.
    */
   public boolean isUnknown()
   {
      return (value == null);
   }
   
   /**
    * Sets this instance to represent the unknown value.
    * <p>
    * Any existing value is discarded.
    */
   public void setUnknown()
   {
      value = null;
   }
   
   /**
    * Sets the data of this instance based on the given instance.
    *
    * @param   value
    *          The instance from which to copy state.
    * @param   force
    *          Flag to indicate if a pending error should be ignored (see 
    *          {@link ErrorManager#isPendingError} and the assignment
    *          forced regardless.  Note that normal assignment silently
    *          exits with no change if there is a pending error. If this
    *          flag is <code>true</code>, that check is bypassed and the
    *          assignment will always be done.
    * 
    * @throws  ClassCastException
    *          if <code>value</code> is not an instance of this class.
    */
   public void assign(BaseDataType value, boolean force)
   {
      assign((rowid) value, force);
   }
   
   /**
    * Sets the data of this instance based on the given instance.
    *
    * @param   value
    *          The instance from which to copy state.
    * @param   force
    *          Flag to indicate if a pending error should be ignored (see 
    *          {@link ErrorManager#isPendingError} and the assignment
    *          forced regardless.  Note that normal assignment silently
    *          exits with no change if there is a pending error. If this
    *          flag is <code>true</code>, that check is bypassed and the
    *          assignment will always be done.
    */
   public void assign(rowid value, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         this.value = value.value;
      }
   }
   
   /**
    * Sets the data of this instance based on the given instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param   value
    *          The instance from which to copy state.
    */
   public void assign(rowid value)
   {
      assign(value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param   value
    *          The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((rowid) value, false);
   }
   
   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The rowids' internal, 64-bit integer values are compared.
    *
    * @param   obj
    *          The instance to compare against.
    *
    * @return   0 or 1 depending on if the passed instance is or is not
    *           <code>unknown</code>.
    */
   public int compareTo(Object obj)   
   {  
      rowid that = null;
      
      if (obj instanceof rowid)
      {
         that = (rowid) obj;
      }
      else
      {
         return -1;
      }
      
      Long l1 = this.value;
      Long l2 = that.value;
      
      if (l1 == l2)
      {
         return 0;
      }
      else if (l1 == null)
      {
         return 1;
      }
      else if (l2 == null)
      {
         return -1;
      }
      
      return l1.compareTo(l2);
   }
   
   /**
    * Creates a string representation of the instance data as a hex string
    * with a leading "0x" prefix.
    * <p>
    * If the instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return  The formatted string.
    */
   public String toString()   
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data as a hex string
    * with a leading "0x" prefix.
    * <p>
    * If the instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @param   fmt
    *          The Progress 4GL format string or <code>null</code> if the
    *          default format is to be used.  This parameter is ignored.
    *
    * @return  The formatted string.
    */
   public String toString(String fmt)   
   {
      if (isUnknown())
      {
         return "?";
      }
      
      StringBuilder buf = new StringBuilder("0x");
      buf.append(Long.toHexString(value).toUpperCase());
      
      return buf.toString();
   }
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents the <code>unknown value</code>, a '?' will be
    * returned.
    *
    * @return  The 'message' formatted string.
    */
   public String toStringMessage()   
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  If the instance represents the <code>unknown value</code>, a
    * '?' will be returned.
    *
    * @return  The 'export' formatted string.
    */
   public String toStringExport()   
   {
      return toString(null);
   }   
   
   /**
    * Return the default display format string for this type.  This
    * implementation returns <code>null</code>, because there is no formatted
    * representation of a rowid:  it must be converted to a string first using
    * the Progress STRING() function.
    *
    * @return  <code>null</code>.
    */
   public String defaultFormatString()   
   {
      return null;
   }
      
   /**
    * Calculate the length of a formatted value (the text form of the value)
    * using this format string.
    *
    * @param    fmt
    *           The format string.
    *
    * @return   The length in characters of the any resulting formatted
    *           value.
    */
   public int formatLength(String fmt)
   {
      return 18;  // an 8-byte hex string + leading '0x' (0x0123456789ABCDEF)
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param   in
    *          The input source from which fields will be restored.
    *
    * @throws  IOException
    *          in case of I/O errors.
    * @throws  ClassNotFoundException
    *          if payload can't be instantiated.
    * @throws  NotSerializableException
    *          always.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      throw new NotSerializableException("rowid");
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param   out
    *          The output destination to which fields will be saved.
    *
    * @throws  IOException
    *          in case of I/O errors.
    * @throws  NotSerializableException
    *          always.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      throw new NotSerializableException("rowid");
   }
}
